import requests
import json
from time import gmtime, strftime
import pymsteams
from soteria.handlers.db import db


class Comms():

    def __init__(self, env='test', teams_url=None, add_urls=None):

        self.env = env
        self.default_url = teams_url

        self.add_urls = add_urls

    def post_message(self, original_test_name, message, fixed=False, message_override=False):
        """ """
        test_name = original_test_name.replace('_', ' ')
        if message_override:
            message = message
        else:
            message = f"**{test_name}:** {message}"

        self.SendTeamsMessage(message, self.default_url)

    def SendTeamsMessage(self, message, webhook):

        print(f"teams message: '{message.replace('*', '')}'")
        myTeamsMessage = pymsteams.connectorcard(webhook)

        myTeamsMessage.text(message)
        try:
            myTeamsMessage.send()
        except requests.exceptions.MissingSchema:
            pass



import time
from datetime import datetime, timedelta, date
from os.path import dirname, basename, isfile, join
import glob
import importlib
import traceback
import sys
import threading
import holidays
import os


from soteria.handlers.db import db
from soteria.handlers.logging import Log
from soteria.handlers.jira import jira
from soteria.module import Module
from soteria.handlers.comms import Comms
from soteria.handlers.scheduler import should_job_run_now
from config import teams_url, run_length, rerun_failed_jobs
try:
    from jobs import *
except Exception as e:
    comms = Comms(env='live', teams_url=teams_url)
    comms.post_message('General', traceback.format_exc())
    sys.exit('unable to import modules')


class SoteriaScheduler:

    def __init__(self, env='test', run_once=False) -> None:
        self.db = db()
        self.env = env
        self.run_once = run_once

        self.comms = Comms(env=self.env,
                           teams_url=teams_url,
                           add_urls='')

        self.jira = jira()

        self.logger = Log()

    def start_scheduler(self):

        # if config set to re-run failed jobs
        if rerun_failed_jobs:
            failed_jobs = self.db.get_failed_jobs()
            for job in failed_jobs:
                self.db.add_to_queue(job)

        try:
            t_end = time.time() + run_length
            while time.time() < t_end:
                self.load_jobs()

                self.update_queue()

                self.run_queue()

                self.db.clear_logs()

                if self.run_once:
                    break

        except KeyboardInterrupt:
            print('Force Closing')

            self.db.clear_queue()
            self.db.set_jobs_to_not_running()

    def run_queue(self):
        """Loops through backlog, running each job"""

        queue = list(set(self.db.get_queue()))

        for job_name in queue:
            # thread = threading.Thread(
            #     target=self.run_job, args=(job_name), kwargs={})
            # thread.start()
            self.run_job(job_name)

    def update_queue(self):

        existing_jobs = self.db.get_jobs_info()

        for job_name in self.jobs:
            job_record = [
                j for j in existing_jobs if j['job_name'] == job_name][0]

            if type(job_record['last_run']) == str:
                last_run = datetime.strptime(
                    job_record['last_run'], '%Y-%m-%d %H:%M:%S')
            else:
                last_run = None

            if should_job_run_now(last_run=last_run, schedule=job_record['schedule']) and self.allowed_to_run(self.jobs[job_name]):
                    self.db.add_to_queue(job_name)

    def run_job(self, job_name):
        """"""
        self.logger.log(f'running: {job_name}')
        
        self.db.update_job_status(job_name=job_name, running=True)

        start_time = time.time()
        try:
            job = globals()[job_name].Job()
        except:
            # Test failed to load
            print(traceback.format_exc())

        try:
            response = job.__run__()
        except:
            # Test failed to run
            print(traceback.format_exc())
            formatted_error = traceback.format_exc(
            ).splitlines()[-1] + "."

            response = {
                'message': f"Test failed to run successfully - { formatted_error }",
                'result': 'error',
            }

        self.action_on_result(
            job_name, result=response['result'], message=response['message'], job=job)

        self.db.update_result(job_name=job_name,
                              result=response['result'],
                              message=response['message'],
                              run_time=int(time.time() - start_time))

        self.db.delete_from_queue(job_name)
        self.db.update_job_status(job_name=job_name, running=False)

    def load_jobs(self):
        """Find all jobs"""

        self.jobs = {}
        file_names = []

        for (dirpath, dirnames, filenames) in os.walk('jobs'):
            # file_names += [os.path.join(dirpath, file) for file in filenames if '.py' in file]
            # print(filenames)
            for file_name in filenames:
                if '.py' in file_name and '.pyc' not in file_name:
                    file_names.append(file_name)
        
        for f in file_names:
            if not '__init__.py' in f:
                self.jobs[basename(f)[:-3]] = globals()[basename(f)[:-3]].Job()

        existing_jobs = self.db.get_jobs_info()
        existing_job_names = [job['job_name'] for job in existing_jobs]

        # add new jobs
        for job_name in self.jobs:
            job = self.jobs[job_name]

            if job_name not in existing_job_names:  # add any new jobs into
                try:
                    self.db.add_job(job_name=job_name,
                                    status=job.setting_enabled,
                                    schedule='n/a')
                except KeyError:
                    # Issue with getting file contents
                    self.db.add_job(job_name=job_name,
                                    status=False,
                                    schedule=job.setting_schedule)
                    self.db.update_job_status(job_name=job_name,
                                              status='error')

            else:
                job = globals()[job_name].Job()
                self.db.job_refresh(job_name=job_name,
                                    status=job.setting_enabled,
                                    schedule=job.setting_schedule)

                # reload job
                try:
                    importlib.reload(globals()[job_name])
                except:
                    # Test failed to complete
                    formatted_error = traceback.format_exc(
                    ).splitlines()[-1] + "."
                    self.db.update_result(
                        job_name=job_name, result='error', message=f'Failed to import. {formatted_error}')

        # delete older jobs
        [self.db.delete_job(job['job_name'])
         for job in existing_jobs if job['job_name'] not in self.jobs]

    def get_jobs_settings(self, job='*'):
        """Returns all the settings for each module"""

        self.load_jobs()

        job_settings = {}
        defaults = self.get_default_variables()

        if job == '*':
            jobs = self.jobs
        else:
            jobs = [job]

        # where a setting has been updated, use that instead of default
        for job_name in jobs:

            current_settings = defaults.copy()
            manual_settings = globals()[job_name].Job().__dict__

            override_settings = {}

            for key in manual_settings:
                override_settings[key.replace(
                    'setting_', '')] = manual_settings[key]

            for setting in current_settings:
                if setting in override_settings:
                    current_settings[setting] = override_settings[setting]

            job_settings[job_name] = current_settings
        return job_settings[job]

    def get_default_variables(self, clean_settings=True):

        defaults = {}
        # grab default settings
        for setting in Module.__dict__:
            if 'setting_' in setting:
                if clean_settings:
                    defaults[setting.replace('setting_', '')] = Module.__dict__[
                        setting]
                else:
                    defaults[setting] = Module.__dict__[
                        setting]
        return defaults

    def get_job_variables(self, job_name):

        secret_values = ['password', 'secret', 'passwd',
                         'authorization', 'api_key', 'apikey', 'access_token', 'username']

        all_variables = globals()[job_name].Job().__dict__

        job_variables = {}
        defaults = self.get_default_variables()

        for setting in all_variables:

            if setting.replace('setting_', '') not in defaults:
                for value in secret_values:
                    job_variables[setting] = all_variables[setting]
                    if value in setting or value in str(all_variables[setting]):
                        job_variables[str(setting)] = '******'
                        break

        return job_variables

    def action_on_result(self, job_name=None, result=None, message=None, job=None):
        """Identifies what to do with the result from the test, acting accordingly."""

        last_result = self.db.get_last_result(job_name)

        if result == 'error':
            result = 'failed'

        if not last_result:
            last_result = 'passed'
        if last_result == 'error':
            last_result = 'failed'

        # PASSED ACTIONS
        if result == 'passed':
            if last_result == 'failed' and job.setting_teams_alerts:
                message = 'Issue has been resolved.'
                self.comms.post_message(job_name, message)

        # FAILED ACTIONS
        elif result == 'failed':
            if last_result == 'passed':
                if job.setting_teams_alerts:
                    self.comms.post_message(job_name, message)
                if job.setting_raise_tickets:
                    self.jira.raise_ticket(job_name, message)
            if last_result == 'failed' and job.setting_teams_alerts:
                # message has changed
                if message != self.db.get_last_message(job_name) and job.setting_teams_alerts:
                    self.comms.post_message(job_name, message)

    def allowed_to_run(self, job=None):

        dayLookup = {"Monday": 0, "Tuesday": 1, "Wednesday": 2, "Thursday": 3, "Friday": 4, "Saturday": 5, "Sunday": 6}
        now = str( datetime.now().time() )
        # Enabled

        if not job.setting_enabled:
            # self.logger.log(f'Job is not enabled')
            return False

        # Bank Holiday Check
        if date.today() in holidays.England() and not job.setting_bank_holidays:
            self.logger.log(f'Job not allowed to run on a bank holiday.')
            return False

        # Time Check
        if now < job.setting_start_time or now > job.setting_end_time:
            self.logger.log(f'Job set to run between {job.setting_start_time} and {job.setting_end_time}.')
            return False

        # Day Check
        for day in job.setting_exception_days:
            if date.today().weekday() == dayLookup[day]:
                self.logger.log(f'Job not set to run on {day}.')
                return False
        
        return True
        

if __name__ == "__main__":

    scheduler = SoteriaScheduler()
    print('starting')
    scheduler.load_jobs()
    # print(scheduler.get_jobs_settings())



from jira import JIRA
from config import jira_url, jira_username, jira_password


class jira:

    def __init__(self, username=None, password=None, url=None):
        try:
            self.jira = JIRA(basic_auth=(jira_username, jira_password),
                             options={'server': jira_url})
            self.enabled = True
            print('jira: enabled')
        except:
            print('jira: disabled (could not connect)')
            self.enabled = False

    def raise_ticket(self, system, message):

        print('RAISING A TICKET')

        issue_dict = {
            'project': {'id': 10001},
            'summary': f'PSS Alert: {system}',
            'description': message,
            'issuetype': {'name': 'Incident'},
            "customfield_12200":  [{"key": "SC-21303"}],
            "customfield_14503":  [{"key": "SC-21669"}]
        }

        if self.enabled:
            ticket_number = self.jira.create_issue(fields=issue_dict)
        else:
            print('Unable to connect to Jira.')
            ticket_number = ''
        return ticket_number



from logging.handlers import TimedRotatingFileHandler
from logging import Formatter
import logging


class Log:

    def __init__(self):

        # get named logger
        self.logger = logging.getLogger(__name__)

        # create handler
        handler = TimedRotatingFileHandler(
            filename='log.log', when='D', interval=1, backupCount=10, encoding='utf-8', delay=False)

        # create formatter and add to handler
        formatter = Formatter(
            fmt='%(asctime)s - %(levelname)s - %(message)s')
        handler.setFormatter(formatter)

        # add the handler to named logger
        self.logger.addHandler(handler)

        # set the logging level
        self.logger.setLevel(logging.INFO)

        # --------------------------------------

        # log something

    def log(self, message):

        print(f'{message}')
        self.logger.info(message)



import sqlite3
from datetime import datetime
from contextlib import closing
import os
import yaml
import base64
import ast

from config import logs_persistence, db_location


class db:

    def __init__(self, startup=False, env='job'):

        self.env = env

        # check if db already exists
        try:
            self.run_query("SELECT * FROM log")
            if startup:
                print('db: connected to existing')
        except sqlite3.OperationalError:
            if startup:
                print('db: not set up, creating now')
            self.build_db()

    def run_query(self, query, params=[], return_dict=False, return_columns=False):
        if self.env == 'live':
            connection = sqlite3.connect(f"{db_location}\log.db")
        else:
            connection = sqlite3.connect(f"{db_location}\log_test.db")
        cursor = connection.cursor()
        rows = cursor.execute(query, params).fetchall()

        if return_dict:
            to_return = []
            columns = [description[0]
                       for description in cursor.description]

            for row in rows:
                new_row = {}
                for count, cell in enumerate(row):
                    new_row[columns[count]] = cell
                to_return.append(new_row)
        elif return_columns:
            return {
                'columns': [description[0]
                            for description in cursor.description],
                'rows': rows
            }
        else:
            to_return = rows

        connection.commit()

        return to_return

    def build_db(self):

        self.run_query(
            "CREATE TABLE log(session_id INTEGER, job_name TEXT, message TEXT, time_logged TEXT)")

        self.run_query(
            "CREATE TABLE results(id INTEGER, job_name TEXT, message TEXT, result TEXT, time_logged TEXT)")

        self.run_query(
            "CREATE TABLE comms(id INTEGER, job_name TEXT, time_logged TEXT)")

        self.run_query(
            "CREATE TABLE queue(job_name TEXT)")

        self.run_query(
            "CREATE TABLE jobs(job_name TEXT, schedule REAL, last_run DATETIME, run_time INTEGER, last_result TEXT, last_message TEXT, running_now TEXT, active TEXT)")

        self.run_query(
            "CREATE TABLE job_history(job_name TEXT, run_completed DATETIME, run_time INTEGER, result TEST, message TEXT)")

    # JOBS

    def get_jobs_info(self, date_type='utc'):
        """Retrives the last message recorded for the job"""

        if date_type.lower() == 'utc':
            query = """SELECT j.job_name,
                    j.schedule,
                    last_run,
                    time(j.run_time, 'unixepoch') [run_time],
                    j.last_result,
                    j.last_message,
                    CASE (j.running_now) WHEN 0 THEN "no" ELSE "yes" END [running_now],
                    CASE (j.active) WHEN 0 THEN "no" ELSE "yes" END [active],
                    CASE (COUNT(q.job_name)) WHEN 0 THEN "no" ELSE "yes" END [in_queue]
                    FROM jobs j
                    LEFT JOIN queue q ON q.job_name=j.job_name
                    GROUP BY j.job_name, j.last_run, j.last_result, j.running_now, j.active """
        elif date_type.lower() == 'local':
            query = """SELECT j.job_name,
                    j.schedule,
                    datetime(last_run, 'localtime') [last_run],
                    time(j.run_time, 'unixepoch') [run_time],
                    j.last_result,
                    j.last_message,
                    CASE (j.running_now) WHEN 0 THEN "no" ELSE "yes" END [running_now],
                    CASE (j.active) WHEN 0 THEN "no" ELSE "yes" END [active],
                    CASE (COUNT(q.job_name)) WHEN 0 THEN "no" ELSE "yes" END [in_queue]
                    FROM jobs j
                    LEFT JOIN queue q ON q.job_name=j.job_name
                    GROUP BY j.job_name, j.last_run, j.last_result, j.running_now, j.active """

        results = self.run_query(query, return_dict=True)

        return results

    def get_job_history(self, job_name):

        query = f"""SELECT job_name,
                           datetime(run_completed, 'localtime') [run_completed],
                           time(run_time, 'unixepoch') [run_time],
                           result,
                           message
                FROM job_history
                WHERE job_name=?
                ORDER BY run_completed DESC"""

        return self.run_query(query, params=[job_name], return_dict=True)

    def add_job(self, job_name=None, status=False, schedule=None, run_time=0):
        print(f'ADDING JOB - {job_name}')

        query = f"INSERT INTO jobs VALUES(?, ?, NULL, ?, 'n/a', 'n/a', ?, ?)"

        self.run_query(
            query, params=[job_name, schedule, run_time, False, status])

    def delete_job(self, job_name=None):
        print(f'REMOVING JOB - {job_name}')

        query = f"DELETE FROM jobs WHERE job_name=?"

        self.run_query(query, params=[job_name])

        query = f"DELETE FROM job_history WHERE job_name=?"

        self.run_query(query, params=[job_name])

    def update_job_status(self, job_name=None, running=False):

        query = f"UPDATE jobs SET running_now=? WHERE job_name=?"

        self.run_query(query, params=[running, job_name])

    def set_jobs_to_not_running(self):
        query = f"UPDATE jobs SET running_now={False}"

        self.run_query(query)

    def update_result(self, job_name=None, result=False, schedule='n/a', message='n/a', run_time=0):

        if run_time == 0:
            run_time = 1

        query = f"UPDATE jobs SET last_result='{result}', last_message= ?, last_run=DATETIME(), run_time={run_time} WHERE job_name='{job_name}'"

        self.run_query(query,  params=(message,))

        query = f"INSERT INTO job_history VALUES ('{job_name}', DATETIME(), {run_time}, '{result}', ?)"

        self.run_query(query,  params=(message,))

    def job_refresh(self, job_name=None, status='', schedule=''):
        query = f"""UPDATE jobs
                    SET active=?,
                        schedule=?
                    WHERE job_name=?"""

        self.run_query(
            query, params=[status, str(schedule), job_name])

    def add_to_queue(self, job_name):

        query = f"INSERT INTO queue VALUES('{job_name}')"

        self.run_query(query)

    # QUEUE

    def get_queue(self):

        query = f"SELECT * FROM queue"

        jobs = [job[0] for job in self.run_query(query)]

        return jobs

    def get_jobs_in_queue(self):

        query = f"""SELECT * FROM queue q
                    LEFT JOIN jobs j ON j.job_name=q.job_name"""

        return self.run_query(query)

    def delete_from_queue(self, job_name=None):
        if job_name == '*':
            query = f"DELETE FROM queue"
        else:
            query = f"DELETE FROM queue WHERE job_name='{job_name}'"

        self.run_query(query)

    def clear_queue(self):
        query = f"DELETE FROM queue"

        self.run_query(query)

    def update_queue(self):

        query = f"SELECT job_name FROM jobs where (next_run<DATETIME() or next_run IS NULL or last_run IS NULL) and active=1"

        jobs = [job[0] for job in self.run_query(query)]

        for job in jobs:
            query = f"INSERT INTO queue VALUES (?)"
            self.run_query(query, params=[job])

    def get_last_run(self, job_name):

        query = f"SELECT last_run FROM jobs where job_name=?"

        last_run = self.run_query(query, params=[job_name], return_dict=True)[
            0]['last_run']

        if last_run == None:
            return None
        else:
            return datetime.strptime(last_run, '%Y-%m-%d %H:%M:%S')

    def get_last_result(self, job_name):

        query = f"SELECT last_result FROM jobs where job_name='{job_name}'"

        return self.run_query(query, return_dict=True)[0]['last_result']

    def get_last_message(self, job_name):

        query = f"SELECT last_message FROM jobs where job_name='{job_name}'"

        return self.run_query(query, return_dict=True)[0]['last_message']

    def clear_logs(self):

        query = f"DELETE FROM job_history WHERE run_completed < datetime('now', '-{str(logs_persistence)} Hour')"

        self.run_query(query)

    def get_failed_jobs(self, provide_message=False):

        if provide_message:
            query = f"SELECT job_name, last_message FROM jobs where last_result IN ('failed', 'error')"

            result = self.run_query(query, return_dict=True)

            return result
        else:

            query = f"SELECT job_name FROM jobs where last_result IN ('failed', 'error')"
            result = self.run_query(query, return_dict=True)

            return [job['job_name'] for job in result]

        return True

    def get_soteria_last_run(self):

        query = f"SELECT last_run FROM jobs ORDER BY last_run DESC LIMIT 1"

        result = self.run_query(query)[0][0]

        return result


if __name__ == "__main__":

    database = db()

    jobs = database.get_failed_jobs()

    print(jobs)



import croniter
import datetime
from datetime import datetime, timedelta


def should_job_run_now(last_run=None, schedule=None):

    if not last_run:
        return True

    cron = croniter.croniter(schedule, last_run)
    next_run = cron.get_next(datetime)

    return next_run < datetime.utcnow()


if __name__ == "__main__":
    sched = '*/5 * * * *'
    now = datetime.utcnow()
    last_ran = now + timedelta(hours=-3)
    print(should_job_run_now(last_run=last_ran, schedule=''))



import yaml
import sys
import base64


class Encryption:

    def __init__(self):

        self.decrypted_values = {}

    def reload_values(self):

        # Load YAML file
        encrypted_values = self.load_yaml()

        # Check for any non-encrypted values, if any found, encrypt them
        if encrypted_values:
            for key in encrypted_values:
                encrypted = encrypted_values[key]

                if type(encrypted) == str:
                    encrypted = self.encrypt_value(encrypted)
                    encrypted_values[key] = encrypted

                self.decrypted_values[key] = self.decrypt_value(encrypted)

            # Save back to file
            self.save_to_file(encrypted_values)

    def encrypt_value(self, value):

        encrypted = base64.urlsafe_b64encode(value.encode("utf-8"))

        return encrypted

    def decrypt_value(self, value):

        return base64.urlsafe_b64decode(value).decode("utf-8")

    def save_to_file(self, values):

        with open('encrypted.yaml', 'w') as f:
            yaml.dump(values, f, width=2000)

    def load_yaml(self):

        try:
            with open('encrypted.yaml', 'r') as stream:
                return yaml.safe_load(stream)
        except FileNotFoundError:
            sys.exit('encrypted.yaml file not found')

    def get_value(self, key):

        self.reload_values()

        try:
            decrypted_value = self.decrypted_values[key]
        except KeyError:
            decrypted_value = 'Key not found in encryption.yaml'

        return decrypted_value



from setuptools import setup, Extension
import os

# read the contents of your README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'Readme.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='soteria',
    packages=['soteria', 'soteria.handlers'],
    version='0.2.1',
    license='MIT',
    description='Scheduling tool for system monitoring.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Euan Campbell',
    author_email='dev@euan.app',
    url='https://github.com/euanacampbell/soteria',
    download_url='https://github.com/euanacampbell/soteria/archive/refs/heads/master.tar.gz',
    keywords=['schedule', 'task'],
    include_package_data=True,
    install_requires=[
        'holidays',
        'PyMySQL',
        'PyYAML',
        'flask',
        'pyodbc',
    ],
    classifiers=[
        # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
)
